#!/usr/bin/env python
"""
Modelica parse Tree to AST tree.
"""
from __future__ import print_function, absolute_import, division, print_function, unicode_literals

import antlr4
import antlr4.Parser
import yaml

from . import ast
from .generated.ModelicaLexer import ModelicaLexer
from .generated.ModelicaListener import ModelicaListener
from .generated.ModelicaParser import ModelicaParser


class ASTListener(ModelicaListener):

    def __init__(self):
        self.ast = {}
        self.ast_result = None
        self.file_node = None
        self.class_node = None
        self.comp_clause_node = None
        self.eq_sect = None
        self.symbol_node = None
        self.eq_comment = None

    # FILE ===========================================================

    def enterStored_definition(self, ctx):
        within = ''
        if ctx.WITHIN() != None:
            within = ctx.WITHIN().getText()
        file_node = ast.File()
        file_node.within = within
        self.ast[ctx] = file_node
        self.file_node = file_node

    def exitStored_definition(self, ctx):
        for class_node in [self.ast[e] for e in ctx.stored_definition_class()]:
            self.ast[ctx].classes[class_node.name] = class_node
        self.ast_result = self.ast[ctx]

    # CLASS ===========================================================

    def enterStored_definition_class(self, ctx):
        class_node = ast.Class()
        class_node.final = ctx.FINAL() != None
        self.class_node = class_node
        self.ast[ctx] = class_node

    def exitStored_definition_class(self, ctx):
        cls = self.class_node
        for sym_key, sym in cls.symbols.items():
            if ((sym.name not in cls.inputs) and
                    (sym.name not in [c.name for c in cls.outputs]) and
                    (sym.name not in [c.name for c in cls.states]) and
                    (sym.name not in [c.name for c in cls.constants]) and
                    (sym.name not in [c.name for c in cls.parameters])):
                cls.variables +=  [ast.ComponentRef(
                    name=sym.name
                )]

    def enterClass_definition(self, ctx):
        class_node = self.class_node
        class_node.encapsulated = ctx.ENCAPSULATED() != None
        class_node.partial = ctx.class_prefixes().PARTIAL() != None
        class_node.type = ctx.class_prefixes().class_type().getText()

    def enterClass_spec_comp(self, ctx):
        class_node = self.class_node
        class_node.name = ctx.IDENT()[0].getText()

    def exitClass_spec_comp(self, ctx):
        class_node = self.class_node
        class_node.comment = self.ast[ctx.string_comment()]

    def exitComposition(self, ctx):
        elist_comb = []
        for elist in [ctx.epriv, ctx.epub, ctx.epro]:
            if elist is not None:
                for e in [self.ast[e] for e in elist.element()]:
                    for s in e.symbol_list:
                        self.class_node.symbols[s.name] = s

        for eqlist in [self.ast[e] for e in ctx.equation_section()]:
            if eqlist is not None:
                if eqlist.initial:
                    self.class_node.initial_equations += eqlist.equations
                else:
                    self.class_node.equations += eqlist.equations

    def enterEquation_section(self, ctx):
        eq_sect = ast.EquationSection(
            initial=ctx.INITIAL() != None
        )
        self.ast[ctx] = eq_sect
        self.eq_sect = eq_sect


    def exitEquation_section(self, ctx):
        eq_sect = self.ast[ctx]
        if eq_sect.initial:
            eq_sect.equations += [self.ast[e] for e in ctx.equation()]
        else:
            eq_sect.equations += [self.ast[e] for e in ctx.equation()]

    # EQUATION ===========================================================

    def enterEquation(self, ctx):
        pass

    def exitEquation(self, ctx):
        self.ast[ctx] = self.ast[ctx.equation_options()]
        try:
            self.ast[ctx].comment = self.ast[ctx.comment()]
        except AttributeError:
            pass

    def exitEquation_simple(self, ctx):
        self.ast[ctx] = ast.Equation(
            left=self.ast[ctx.simple_expression()],
            right=self.ast[ctx.expression()])

    def exitEquation_connect_clause(self, ctx):
        self.ast[ctx] = self.ast[ctx.connect_clause()]

    def exitConnect_clause(self, ctx):
        self.ast[ctx] = ast.ConnectClause(
            left= self.ast[ctx.component_reference()[0]],
            right= self.ast[ctx.component_reference()[1]])

    # EXPRESSIONS ===========================================================

    def exitSimple_expression(self, ctx):
        #TODO only using first expression
        self.ast[ctx] = self.ast[ctx.expr()[0]]

    def exitExpression_simple(self, ctx):
        self.ast[ctx] = self.ast[ctx.simple_expression()]

    def exitExpr_primary(self, ctx):
        self.ast[ctx] = self.ast[ctx.primary()]

    def exitExpr_add(self, ctx):
        self.ast[ctx] = ast.Expression(
            operator=ctx.op.text,
            operands=[self.ast[e] for e in ctx.expr()]
        )

    def exitExpr_mul(self, ctx):
        self.ast[ctx] = ast.Expression(
            operator=ctx.op.text,
            operands=[self.ast[e] for e in ctx.expr()]
        )

    def exitExpr_rel(self, ctx):
        self.ast[ctx] = ast.Expression(
            operator=ctx.op.text,
            operands=[self.ast[e] for e in ctx.expr()]
        )

    def exitExpr_neg(self, ctx):
        self.ast[ctx] = ast.Expression(
            operator=ctx.op.text,
            operands=[self.ast[ctx.expr()]]
        )


    # PRIMARY ===========================================================

    def exitPrimary_unsigned_number(self, ctx):
        self.ast[ctx] = ast.Primary(value=str(yaml.load(ctx.getText())))

    def exitPrimary_string(self, ctx):
        self.ast[ctx] = ast.Primary(value=ctx.getText())

    def exitPrimary_false(self, ctx):
        self.ast[ctx] = ast.Primary(value=False)

    def exitPrimary_true(self, ctx):
        self.ast[ctx] = ast.Primary(value=True)

    def exitPrimary_function(self, ctx):
        self.ast[ctx] = ast.Primary(value=ctx.getText())

    def exitPrimary_derivative(self, ctx):
        self.ast[ctx] = ast.Primary(value=ctx.getText())
        comp_name =  ctx.function_call_args().function_arguments().function_argument()[0].getText()
        self.ast[ctx] = ast.Expression(
            operator='der',
            operands=[ast.ComponentRef(name=comp_name)]
        )

        # every symbols that is differentiated is a state
        # make sure it isn't already a state
        for state in self.class_node.states:
            if state == comp_name:
                return
        self.class_node.states += [ast.ComponentRef(name=comp_name)]

    def exitComponent_reference(self, ctx):
        # TODO handle other idents
        self.ast[ctx] = ast.ComponentRef(
            name=ctx.IDENT()[0].getText()
        )

    def exitPrimary_component_reference(self, ctx):
        self.ast[ctx] = ast.ComponentRef(
            name=ctx.getText()
        )

    def exitEquation_function(self, ctx):
        # TODO, add function ast
        self.ast[ctx] = ctx.getText()

    def exitEquation_when(self, ctx):
        # TODO, add when ast
        self.ast[ctx] = ctx.getText()

    # COMPONENTS ===========================================================

    def exitElement_list(self, ctx):
        self.ast[ctx] = [self.ast[e] for e in ctx.element()]

    def exitElement(self, ctx):
        self.ast[ctx] = self.ast[ctx.regular_element()]

    def exitRegular_element(self, ctx):
        self.ast[ctx] = self.ast[ctx.comp_elem]

    def enterComponent_clause(self, ctx):
        if ctx.array_subscripts() is not None:
            dimensions=[int(s) for s in ctx.array_subscripts().subscript().getText()]
        else:
            dimensions = [1]
        self.ast[ctx] = ast.ComponentClause(
            prefixes=ctx.type_prefix().getText().split(' '),
            type=ctx.type_specifier().getText(),
            dimensions=dimensions
        )
        self.comp_clause = self.ast[ctx]

    def exitComponent_clause(self, ctx):
        self.ast[ctx].symbol_list = self.ast[ctx.component_list()]

    def exitComponent_list(self, ctx):
        self.ast[ctx] = [self.ast[e] for e in ctx.component_declaration()]

    def enterComponent_declaration(self, ctx):
        sym = ast.Symbol(
            name='',
            type=self.comp_clause.type,
            prefixes=self.comp_clause.prefixes,
        )
        self.symbol_node = sym
        self.ast[ctx] = sym

    def exitComponent_declaration(self, ctx):
        self.ast[ctx].comment = self.ast[ctx.comment()]

    def enterDeclaration(self, ctx):
        sym = self.symbol_node
        dimensions = None
        if ctx.array_subscripts() is not None:
            dimensions = [int(s) for s in ctx.array_subscripts().subscript().getText()]
        elif self.comp_clause.dimensions is not None:
            dimensions = self.comp_clause.dimensions
        sym.name = ctx.IDENT().getText()
        sym.dimensions = dimensions
        #if ctx.modification() is not None:
        #    sym.start = self.ast[ctx.modification()]
        if 'input' in sym.prefixes:
            self.class_node.inputs += [ast.ComponentRef(name=sym.name)]
        elif 'output' in sym.prefixes:
            self.class_node.outputs += [ast.ComponentRef(name=sym.name)]
        elif 'constant' in sym.prefixes:
            self.class_node.constants += [ast.ComponentRef(name=sym.name)]
        elif 'parameter' in sym.prefixes:
            self.class_node.parameters += [ast.ComponentRef(name=sym.name)]

    def exitElement_modification(self, ctx):
        sym = self.symbol_node
        if ctx.name().getText() == 'start':
            sym.start = self.ast[ctx.modification().expression()]

    def exitModification_assignment(self, ctx):
        sym = self.symbol_node
        sym.start = self.ast[ctx.expression()]

    # COMMENTS ==============================================================

    def exitComment(self, ctx):
        # TODO handle annotation
        self.ast[ctx] = self.ast[ctx.string_comment()]

    def exitString_comment(self, ctx):
        self.ast[ctx] = ctx.getText()[1:-1]

# UTILITY FUNCTIONS =========================================================

def parse(text):
    antlr4.InputStream
    input_stream = antlr4.InputStream(text)
    lexer = ModelicaLexer(input_stream)
    stream = antlr4.CommonTokenStream(lexer)
    parser = ModelicaParser(stream)
    parse_tree = parser.stored_definition()
    astListener = ASTListener()
    parse_walker = antlr4.ParseTreeWalker()
    parse_walker.walk(astListener, parse_tree)
    ast_tree = astListener.ast_result
    return ast_tree



# THIS FILE IS GENERATED FROM SETUP.PY
short_version = '0.1.4'
version = '0.1.4'
full_version = '0.1.4'
git_revision = '1787df7398bf2b384288f2a7d3d9a7da87044dea'
release = True

if not release:
    version = full_version



#!/usr/bin/env python
"""
Tools for tree walking and visiting etc.
"""

from __future__ import print_function, absolute_import, division, print_function, unicode_literals

from . import ast


class TreeWalker(object):
    def walk(self, listener, tree):
        name = tree.__class__.__name__
        if hasattr(listener, 'enterEvery'):
            getattr(listener, 'enterEvery')(tree)
        if hasattr(listener, 'enter' + name):
            getattr(listener, 'enter' + name)(tree)
        for child_name in tree.ast_spec.keys():
            self.handle_walk(self, listener, tree.__dict__[child_name])
        if hasattr(listener, 'exitEvery'):
            getattr(listener, 'exitEvery')(tree)
        if hasattr(listener, 'exit' + name):
            getattr(listener, 'exit' + name)(tree)

    @classmethod
    def handle_walk(cls, walker, listener, tree):
        if isinstance(tree, ast.Node):
            walker.walk(listener, tree)
        elif isinstance(tree, dict):
            for k in tree.keys():
                cls.handle_walk(walker, listener, tree[k])
        elif isinstance(tree, list):
            for i in range(len(tree)):
                cls.handle_walk(walker, listener, tree[i])
        else:
            pass


class TreeVisitor(object):
    """
    TODO this class hasn't been tested
    """

    def visit(self, visitor, tree):
        name = self.__class__.__name__
        if hasattr(visitor, 'visit' + name):
            return visitor.__getattribute__('visit' + name)(tree)
        else:
            res = []
            for k in tree.ast_spec.keys():
                res += [self.handle_visit(visitor, tree[k])]
            if len(res) == 1:
                res = res[0]
            return res

    @classmethod
    def handle_visit(cls, visitor, tree):
        res = []
        if isinstance(tree, ast.Node):
            res += visitor.visit(visitor, tree)
        elif isinstance(tree, dict):
            for k in tree.keys():
                res += [cls.handle_visit(tree[k], visitor)]
        elif isinstance(tree, list):
            for c in tree:
                res += [cls.handle_visit(c, visitor)]
        if len(res) == 1:
            res = res[0]
        return res


class TreeListener(object):
    def enterFile(self, tree):
        pass

    def exitFile(self, tree):
        pass

    def enterClass(self, tree):
        pass

    def exitClass(self, tree):
        pass

    def enterExpression(self, tree):
        pass

    def exitExpression(self, tree):
        pass

    def exitConnectClause(self, tree):
        pass

    def exitSymbol(self, tree):
        pass

    def exitComponentClause(self, tree):
        pass

    def exitPrimary(self, tree):
        pass

def flatten(root, class_name, instance_name=''):
    """
    This function takes and flattens it so that all subclasses instances
    are replaced by the their equations and symbols with name mangling
    of the instance name passed.
    :param root: The root of the tree that contains all class definitions
    :param class_name: The class we want to flatten
    :param instance_name:
    :return: flat_class, the flattened class of type Class
    """

    # extract the original class of interest
    orig_class = root.classes[class_name]

    # create the returned class
    flat_class = ast.Class(
        name=class_name,
        equations=orig_class.equations,
        states = orig_class.states,
        inputs = orig_class.inputs,
        outputs = orig_class.outputs,
        parameters = orig_class.parameters,
        constants = orig_class.constants,
        variables = orig_class.variables,
    )

    # flat file
    flat_file = ast.File()
    flat_file.classes[class_name] = flat_class

    # append period to non empty instance_name
    if instance_name != '':
        instance_prefix = instance_name + '.'
    else:
        instance_prefix = instance_name

    # for all symbols in the original class
    for sym_name, sym in orig_class.symbols.items():
        # if the symbol type is a class
        if sym.type in root.classes:
            # recursively call flatten on the sub class
            flat_sub_file = flatten(root, sym.type, instance_name=sym_name)
            flat_sub_class = flat_sub_file.classes[sym.type]

            # add sub_class members symbols and equations
            for sub_sym_name, sub_sym in flat_sub_class.symbols.items():
                flat_class.symbols[instance_prefix + sub_sym_name] = sub_sym
            flat_class.equations += flat_sub_class.equations
            flat_class.states += flat_sub_class.states
            #flat_class.inputs += flat_sub_class.inputs
            #flat_class.outputs += flat_sub_class.outputs
            flat_class.variables += flat_sub_class.variables
            flat_class.parameters += flat_sub_class.parameters
            flat_class.constants += flat_sub_class.constants

        # else if the symbols is not a class name
        else:
            # append original symbol to flat class
            flat_class.symbols[instance_prefix + sym_name] = sym


    # walker for renaming components
    if instance_name != '':
        ast_walker = TreeWalker()
        ast_walker.walk(ComponentRenameListener(instance_name), flat_class)

    return flat_file


class ComponentRenameListener(object):
    def __init__(self, prefix):
        self.prefix = prefix

    def enterSymbol(self, tree):
        tree.name = '{:s}.{:s}'.format(self.prefix, tree.name)

    def enterComponentRef(self, tree):
        tree.name = '{:s}.{:s}'.format(self.prefix, tree.name)


#!/usr/bin/env python
"""
Modelica AST definitions
"""
from __future__ import print_function, absolute_import, division, print_function, unicode_literals

import copy
import json
import sys

VALIDATE_AST = True

"""
AST Node Type Hierarchy

File
    Class
        Equation
            ComponentRef
            Expression
            Primary
        ConnectClause
            ComponentRef
        Symbol
"""

def to_json(var):
    if isinstance(var, list):
        res = [to_json(item) for item in var]
    elif isinstance(var, dict):
        res = {key: to_json(var[key]) for key in var.keys()}
    elif isinstance(var, Node):
        res = {key: to_json(var.__dict__[key]) for key in var.__dict__.keys()}
    else:
        res = var
    return res


class Field(object):
    def __init__(self, types, default=None):
        if default == None:
            if types == dict:
                default = {}
            elif types == list:
                default = []
        if type(types) is type or not hasattr(types, '__iter__'):
            types = [types]
        types = list(types)
        if sys.version_info < (3,):
            if str in types:
                types += [unicode]
        self.types = types
        self.default = default

    def validate(self, name, key, val, throw=True):
        if not type(val) in self.types:
            if throw:
                raise IOError('{:s}.{:s} requires types {:s}, but got {:s}'.format(
                    name, key, self.types, type(val)))
            else:
                return False
        else:
            return True


# TODO, need to finish implementing this for list type checking
class List(list):
    def __init__(self, types):
        super(List, self).__init__()
        if not isinstance(types, list):
            types = list([types])
        self.types = types

    def __add__(self, other):
        if other in self.types:
            super(List, self).__add__(other)
        else:
            raise IOError('List requires elements of type', self.types)


class Node(object):
    ast_spec = {}

    def __init__(self, **kwargs):
        for key in self.ast_spec.keys():
            # make sure we don't share ast_spec default data by using deep copy
            self.__dict__[key] = copy.deepcopy(self.ast_spec[key].default)
        for key in kwargs.keys():
            types = self.ast_spec[key].types
            val = kwargs[key]
            self.set_field(key, val)

    def set_field(self, key, value):
        if VALIDATE_AST:
            name = self.__class__.__name__
            if key not in self.ast_spec.keys():
                raise IOError('{:s} not a child of {:s}'.format(key, name))
            self.ast_spec[key].validate(name, key, value)
        self.__dict__[key] = value

    def __setattr__(self, key, value):
        self.set_field(key, value)

    def __repr__(self):
        return json.dumps(to_json(self), indent=2, sort_keys=True)

    __str__ = __repr__


class Primary(Node):
    ast_spec = {
        'value': Field((bool, float, int, str))
    }

    def __init__(self, **kwargs):
        super(Primary, self).__init__(**kwargs)


class ComponentRef(Node):
    ast_spec = {
        'name': Field(str),
    }

    def __init__(self, **kwargs):
        super(ComponentRef, self).__init__(**kwargs)


# so expression can reference itself
class Expression(Node):
    pass


# noinspection PyRedeclaration
class Expression(Node):
    ast_spec = {
        'operator': Field(str),
        'operands': Field(list),  # list of Expressions
    }

    def __init__(self, **kwargs):
        super(Expression, self).__init__(**kwargs)


class Equation(Node):
    ast_spec = {
        'left': Field((Expression, Primary, ComponentRef)),
        'right': Field((Expression, Primary, ComponentRef)),
        'comment': Field(str),
    }

    def __init__(self, **kwargs):
        super(Equation, self).__init__(**kwargs)


class ConnectClause(Node):
    ast_spec = {
        'left': Field(ComponentRef),
        'right': Field(ComponentRef),
        'comment': Field(str),
    }

    def __init__(self, **kwargs):
        super(ConnectClause, self).__init__(**kwargs)


class Symbol(Node):
    ast_spec = {
        'name': Field(str, ''),
        'type': Field(str, ''),
        'prefixes': Field(list, []),  # (str)
        'redeclare': Field(bool, False),
        'final': Field(bool, False),
        'inner': Field(bool, False),
        'outer': Field(bool, False),
        'dimensions': Field(list, [1]),  # (int)
        'comment': Field(str, ''),
        'start': Field(Primary, ''),
    }

    def __init__(self, **kwargs):
        super(Symbol, self).__init__(**kwargs)


class ComponentClause(Node):
    ast_spec = {
        'prefixes': Field(list, []),  # (str)
        'type': Field(str, ''),
        'dimensions': Field(list, [1]),  # (int)
        'comment': Field(str, ''),
        'symbol_list': Field(list, []),  # (Symbol)
    }

    def __init__(self, **kwargs):
        super(ComponentClause, self).__init__(**kwargs)


class EquationSection(Node):
    ast_spec = {
        'initial': Field(bool, False),
        'equations': Field(list, []),  # (Equation)
    }

    def __init__(self, **kwargs):
        super(EquationSection, self).__init__(**kwargs)


class Class(Node):
    ast_spec = {
        'name': Field(str),
        'encapsulated': Field(bool, False),
        'partial': Field(bool, False),
        'final': Field(bool, False),
        'type': Field(str, ''),
        'comment': Field(str, ''),
        'symbols': Field(dict, {}),  # (Symbol)
        'initial_equations': Field(list, []),
        'equations': Field(list, []),  # (Equation)
        'parameters': Field(list, []),  # (ComponentRef)
        'constants': Field(list, []),  # (ComponentRef)
        'inputs': Field(list, []),  # (ComponentRef)
        'outputs': Field(list, []),  # (ComponentRef)
        'states': Field(list, []),  # (ComponentRef)
        'variables': Field(list, []),  # (ComponentRef)
    }

    def __init__(self, **kwargs):
        super(Class, self).__init__(**kwargs)


class File(Node):
    ast_spec = {
        'within': Field(str),
        'classes': Field(dict),
    }

    def __init__(self, **kwargs):
        super(File, self).__init__(**kwargs)



from __future__ import print_function, absolute_import, division, print_function, unicode_literals

import jinja2
import os
import sys

FILE_DIR = os.path.dirname(os.path.realpath(__file__))

from . import tree


class SympyGenerator(tree.TreeListener):
    indent='    '

    def __init__(self):
        self.src = {}

    def exitFile(self, tree):
        d = {'classes': []}
        for key in sorted(tree.classes.keys()):
            d['classes'] += [self.src[tree.classes[key]]]

        template = jinja2.Template('''
# do not edit, generated by pymola

from __future__ import print_function, division
import sympy
import sympy.physics.mechanics as mech
from pymola.sympy_runtime import OdeModel

{%- for class_key, class in tree.classes.items() %}
{{ render.src[class] }}
{%- endfor %}
''')
        self.src[tree] = template.render({
            'tree': tree,
            'render': self,
        })

    def exitClass(self, tree):

        states_str = ', '.join([self.src[s] for s in tree.states])
        inputs_str = ', '.join([self.src[s] for s in tree.inputs])
        outputs_str = ', '.join([self.src[s] for s in tree.outputs])
        constants_str = ', '.join([self.src[s] for s in tree.constants])
        parameters_str = ', '.join([self.src[s] for s in tree.parameters])
        variables_str = ', '.join([self.src[s] for s in tree.variables])

        d = locals();
        d.pop('self')
        d['render'] = self

        template = jinja2.Template('''
class {{tree.name}}(OdeModel):

    def __init__(self):

        super({{tree.name}}, self).__init__()

        # states
        {% if states_str|length > 0 -%}
        {{ states_str }} = mech.dynamicsymbols('{{ states_str|replace('__', '.') }}')
        {% endif -%}
        self.x = sympy.Matrix([{{ states_str }}])

        # inputs
        {% if inputs_str|length > 0 -%}
        {{ inputs_str }} = sympy.symbols('{{ inputs_str|replace('__', '.') }}')
        {% endif -%}
        self.u = sympy.Matrix([{{ inputs_str }}])

        # outputs
        {% if outputs_str|length > 0 -%}
        {{ outputs_str }} = sympy.symbols('{{ outputs_str|replace('__', '.') }}')
        {% endif -%}
        self.y = sympy.Matrix([{{ outputs_str }}])

        # constants
        {% if constants_str|length > 0 -%}
        {{ constants_str }} = sympy.symbols('{{ constants_str|replace('__', '.') }}')
        {% endif -%}
        self.c = sympy.Matrix([{{ constants_str }}])
        self.c0 = {
            {% for s in tree.constants -%}
            '{{ s.name }}' : {{tree.symbols[s.name].start.value}},
            {% endfor -%}}

        # parameters
        {% if parameters_str|length > 0 -%}
        {{ parameters_str }} = sympy.symbols('{{ parameters_str|replace('__', '.') }}')
        {% endif -%}
        self.p = sympy.Matrix([{{ parameters_str }}])
        self.p0 = {
            {% for s in tree.parameters -%}
            '{{ s.name }}' : {{tree.symbols[s.name].start.value}},
            {% endfor -%}}

        # variables
        {% if variables_str|length > 0 -%}
        {{ variables_str }} = sympy.symbols('{{ variables_str|replace('__', '.') }}')
        {% endif -%}
        self.v = sympy.Matrix([{{ variables_str }}])
      
        # equations
        self.eqs = [
            {% for eq in tree.equations -%}
            {{ render.src[eq] }},
            {% endfor -%}
        ]

        self.compute_fg()
    ''')
        self.src[tree] = template.render(d)

    def exitExpression(self, tree):
        op = str(tree.operator)
        n_operands = len(tree.operands)
        if op == 'der':
            src = '({var:s}).diff(self.t)'.format(
                var=self.src[tree.operands[0]])
        elif op in ['*', '+', '-', '/'] and n_operands == 2:
            src = '{left:s} {op:s} {right:s}'.format(
                op=op,
                left=self.src[tree.operands[0]],
                right=self.src[tree.operands[1]])
        elif op in ['+', '-'] and n_operands == 1:
            src = '{op:s} {expr:s}'.format(
                op=op,
                expr=self.src[tree.operands[0]])
        else:
            src = "({operator:s} ".format(**tree.__dict__)
            for operand in tree.operands:
                src +=  ' ' + self.src[operand]
            src += ")"
        self.src[tree] = src

    def exitPrimary(self, tree):
        self.src[tree] = "{value:s}".format(**tree.__dict__)

    def exitComponentRef(self, tree):
        self.src[tree] = "{name:s}".format(name=tree.name.replace('.','__'))

    def exitSymbol(self, tree):
        self.src[tree] = "{name:s}".format(name=tree.name.replace('.','__'))

    def exitEquation(self, tree):
        self.src[tree] = "{left:s} - ({right:s})".format(
            left=self.src[tree.left],
            right=self.src[tree.right])

def generate(ast_tree, model_name):
    ast_walker = tree.TreeWalker()
    flat_tree = tree.flatten(ast_tree, model_name)
    sympy_gen = SympyGenerator()
    ast_walker.walk(sympy_gen, flat_tree)
    return sympy_gen.src[flat_tree]
